(*

  application: image filtering
  state: experimental
  revision: $Id: defs.ml,v 1.3 2005/06/28 07:33:08 tobias Exp $

  Metaprogramming Project
  Authors: Christoph A. Herrmann and Tobias Langhammer
  University of Passau, Department of Informatics, Chair of Programming

 Copyright (C) 2004    University of Passau, Department of Informatics

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*)

exception Type_error of string
exception Lookup_error of string


(******** datatypes for abstract syntax tree **************)
type value  =
    VInt of int
  | VFloat of float
  | VBool of bool
  | VMatrix of float array array

type dtype = Bool | Int | Float | Matrix

type unOp
 = NegI | NegF
 | Not
 | AbsI | AbsF
 | Exp | Log | Sqrt | Sin | Cos | Tan | ASin | ACos | ATan

type binOp
 = AddI | SubI | MulI | DivI | PowI | MinI | MaxI | ModI
 | AddF | SubF | MulF | DivF | PowF | MinF | MaxF
 | EqB  | UEqB | AndB | OrB
 | EqI  | UEqI | LtI  | LeI  | GtI  | GeI
 | EqF  | UEqF | LtF  | LeF  | GtF  | GeF

  type op =
      C of value
    | V of string
    | Image
    | Int2Float
    | Floor
    | UnOp  of unOp
    | BinOp of binOp
    | If
    | Let of string
    | IndexMatrix of string
    | Sum of string

  type exp = { dtype:dtype; op:op; args:exp list; expand:int }

  let create_exp dtype op args =
    match args with
    | [] -> { dtype=dtype; op=op; args=[]; expand=1 }
    | _ ->  let add_expand n arg = n + arg.expand in
            let expand = List.fold_left add_expand 0 args in
            { dtype=dtype; op=op; args=args; expand=expand }

  type top_exp = { common_bind: (string * exp) list; 
                   channels: exp list }
  
let exp_of_value = function
  | VBool i -> create_exp Bool (C (VBool i)) []
  | VInt i -> create_exp Int (C (VInt i)) []
  | VFloat i -> create_exp Float (C (VFloat i)) []
  | VMatrix i -> create_exp Matrix (C (VMatrix i)) []



let type_of_binop = function
 |AddI | SubI | MulI | DivI
 | PowI | MinI | MaxI | ModI -> Int
 | AddF | SubF | MulF
 | DivF | PowF | MinF | MaxF -> Float
 | EqB  | UEqB | AndB | OrB -> Bool
 | EqI  | UEqI | LtI  | LeI  | GtI  | GeI
 | EqF  | UEqF | LtF  | LeF  | GtF  | GeF -> Bool

let unC = function
  | { op=C c } -> Some c
  | _          -> None



(* some helpers... *)

let rec size_exp {args=args} =
  1 + List.fold_left (fun x y -> x + size_exp y) 0 args
let size_top_exp {common_bind=bs; channels=chs } =
  1 + List.fold_left (fun x (_,y) -> x + size_exp y) 0 bs
  + List.fold_left (fun x y -> x + size_exp y) 0 chs

let head ?(info="") = function
  | a::_ -> a
  | _ -> raise (Failure
      ("in 'head': Wrong # of many arguments. Bug in Parser?"^info) );;

let head2 = function
  | a::b::_ -> (a,b)
  | _ -> raise (Failure
      "in 'head2': Wrong # of many arguments. Bug in Parser?");;

let head3 = function
  | a::b::c::_ -> (a,b,c)
  | _ -> raise (Failure
      "in 'head3': Wrong # of many arguments. Bug in Parser?");;




(*
 * Functions for converting AST to string.
 * These are mainly for debugging
 *)
let string_of_dtype = function
  | Bool -> "b"
  | Int -> "i"
  | Float -> "f"
  | Matrix -> "m"

let string_of_unOp = function
 | NegI -> "NegI"
 | NegF -> "NegF"
 | Not -> "Not"
 | AbsI -> "AbsI"
 | AbsF -> "AbsF"
 | Exp -> "Exp"
 | Log -> "Log"
 | Sqrt -> "Sqrt"
 | Sin -> "Sin"
 | Cos -> "Cos"
 | Tan -> "Tan"
 | ASin -> "ASin"
 | ACos -> "ACos"
 | ATan -> "ATan"

let string_of_binOp = function
  | AddI -> "AddI"
  | SubI -> "SubI"
  | MulI -> "MulI"
  | DivI -> "DivI"
  | PowI -> "PowI"
  | MinI -> "MinI"
  | MaxI -> "MaxI"
  | ModI -> "ModI"
  | AddF -> "AddF"
  | SubF -> "SubF"
  | MulF -> "MulF"
  | DivF -> "DivF"
  | PowF -> "PowF"
  | MinF -> "MinF"
  | MaxF -> "MaxF"
  | EqB -> "EqB"
  | UEqB -> "UEqB"
  | AndB -> "AndB"
  | OrB -> "OrB"
  | EqI -> "EqI"
  | UEqI -> "UEqI"
  | LtI -> "LtI"
  | LeI -> "LeI"
  | GtI -> "GtI"
  | GeI -> "GeI"
  | EqF -> "EqF"
  | UEqF -> "UEqF"
  | LtF -> "LtF"
  | LeF -> "LeF"
  | GtF -> "GtF"
  | GeF -> "GeF"

let string_of_value = function
  | VInt i -> Printf.sprintf "%d" i
  | VFloat f -> Printf.sprintf "%f" f
  | VBool true -> "true"
  | VBool false -> "false"
  | VMatrix m ->
      let acc = ref "[" in
      for i = 0 to Array.length m-1 do
        acc := !acc ^ if i = 0 then "[" else " ";
        for j = 0 to Array.length m.(0) -1 do
          acc := !acc ^ if j = 0 then "[" else " ";
          acc:= Printf.sprintf "%s%f" !acc m.(i).(j);
        done;
        acc := !acc^"]"
      done;
      acc := ! acc^"]";
      !acc

let string_of_op = function
  | C v -> string_of_value v
  | V s -> Printf.sprintf "\"%s\"" s
  | Image -> "image"
  | Int2Float -> "(float)"
  | Floor -> "floor"
  | UnOp op -> string_of_unOp op
  | BinOp op -> string_of_binOp op
  | If -> "if"
  | Let s -> Printf.sprintf "let %s" s
  | IndexMatrix s -> Printf.sprintf "%s[]" s
  | Sum s -> Printf.sprintf "sum %s" s


let string_of_exp n =
  let rec soe indent { dtype=t; op=o; args=es } =
    let head = Printf.sprintf "%s{%s|%s|"
      indent (string_of_op o)
      (string_of_dtype t)
    in
    let block =
      if es = [] then "}\n" else
        let step exp acc =
          Printf.sprintf "%s%s" (soe (indent^"  ") exp) acc
        in
        "\n" ^ List.fold_right step es "" ^ indent ^ "}\n"
    in
    head ^ block
  in soe "" n

let print_exp e = print_string (string_of_exp e)

(*
let rec fromto a b = if a>b then [] else a :: fromto (a+1) b

let rec fold_range f a lb ub =
  if lb > ub then a
  else
    fold_range f (f a lb) (lb+1) ub

let rec split_at n xs =
  match (n,xs) with
  | (0,rest) -> ([],rest)
  | (n,[]) -> ([],[])
  | (n, x::xs') ->
      let (acc',rest) = split_at (n-1) xs' in (x::acc',rest)

*)
(******** datatype and functions for environment **********)
type 'a environment = string -> 'a option

let extEnv (k,v) env s = if s=k then Some v else env s

let emptyEnv s = None

let newEnv pairs = List.fold_right extEnv pairs emptyEnv
let extEnvPairs pairs env = List.fold_right extEnv pairs env
let lookup n env = env n

let lookup_excp n env =
  match lookup n env with
  | Some v -> v
  | None -> raise (Lookup_error ("symbol "^n^" not in environment"))


let dynamic_env =
  [
    "row",Int,None;
    "col",Int,None
  ]
let static_env =
  [
    "red",Int,Some (VInt 0);
    "green",Int,Some (VInt 1);
    "blue",Int,Some (VInt 2);
    "gray",Int,Some (VInt 3);
    "iter",Int,None;
    "width",Int,None;
    "height",Int,None;
    "maxval",Int,Some (VInt 255)
  ]


