(* vim: ft=ocaml

  application: image filtering
  state: experimental
  revision: $Id: main.ml,v 1.3 2005/06/21 11:39:41 tobias Exp $
 
  Metaprogramming Project
  Authors: Christoph A. Herrmann and Tobias Langhammer
  University of Passau, Department of Informatics, Chair of Programming

 Copyright (C) 2004    University of Passau, Department of Informatics
 
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*)


open Printf
open Filt_config
open Ppm 

let exec_times n (f:unit -> 'a) =
  for i = 1 to n do
    f ()
  done

 
let parse formula = begin
  let lexbuf = Lexing.from_string formula in
  begin try
      Filt_parser.main Filt_lexer.token lexbuf
  with 
  | Parsing.Parse_error ->
    begin
      let pos = lexbuf.Lexing.lex_start_p.Lexing.pos_cnum in
      prerr_string "Parse error in the following expression:\n\n";
      prerr_string lexbuf.Lexing.lex_buffer;
      prerr_string "\nUnexpected character '";
      prerr_char lexbuf.Lexing.lex_buffer.[pos];
      prerr_string "' at position ";
      prerr_int pos;
      prerr_string "!\n";
      prerr_string "\nexiting.\n";
      flush stderr;
      raise Parsing.Parse_error; 
    end
  | Filt_lexer.Eof -> raise (Failure "parse error: EOF");
  end
end
  
  
let swap a b = 
  begin
    let t = !a in
    a := !b;
    b := t;
  end

let runfilter cfg in_img out_img =
  begin
    cfg.prerr 2 (lazy (
     Printf.sprintf "%s;%d;%d;%s;%s\n"
        cfg.label
        ((!out_img).width) 
        ((!out_img).height) 
        (if cfg.staged then "staged" else "unstaged")
        (if cfg.simplify then "simplify" else "no simplify")));
    Timing.init_times ();
    let ast = 
        parse cfg.formula
    (*  Timing.timenew_cond cfg.time "parsing"  (fun () ->
        parse cfg.formula
      *)
    in

    let run_filter filter i = begin
      let top = cfg.margintop in
      let bottom = (!in_img).height-1-cfg.marginbottom in
      let left = cfg.marginleft in
      let right = (!in_img).width-1-cfg.marginright in
      for row = top to bottom do
        for col = left to right do
          Ppm.set_point !out_img row col (filter (row,col))
        done
      done;
    end in
    
    for i = 0 to cfg.iter-1 do
      cfg.prerr 2 (lazy ("iteration "^string_of_int i^"\n"));

      (* STAGE 1: SIMPLIFICATION *)
      let simplifyer () = cfg.simplifyer !in_img i ast in
      let ast2 = Timing.timenew_cond cfg.time_simplify "simplify" simplifyer in

      (* STAGE 2: FILTERING *)
      let filter = cfg.compiler !in_img i ast2 in
      Timing.timenew_cond cfg.time_exec "exec" (fun () -> run_filter filter i);
      begin match cfg.osuffix with
      | Some osuffix ->
          let oc = open_out (Printf.sprintf "%0.5d%s" i osuffix) in
          Ppm.output_ppm oc !out_img;
          close_out oc;
      | None -> ();
      end;
      if not cfg.iter_indep then swap out_img in_img
    done;
    if cfg.iter_indep then swap out_img in_img;

  end
      

(* main *)
let _ =
  try
    let cfg = Filt_config.get () in
    cfg.prerr 2 (lazy (Printf.sprintf "max_unrolling=%d\n" cfg.max_unrolling));
    begin match cfg.mode with
    | Benchmark (i, f) ->
      for k = 0 to i-1 do
        let size = f k in
        cfg.print 2 (lazy (Printf.sprintf "size;%d\n" size));
        let in_img = ref (Ppm.create size size 255 Ppm.black) in
        let out_img = ref (Ppm.create size size 255 Ppm.black) in
        runfilter cfg in_img out_img;
        Timing.print_times ();
      done
    | Normal ->
      let in_img = ref (Ppm.input_ppm cfg.ic) in
      let out_img =
        let { height=h; width=w; maxval=m } = !in_img in 
        ref (Ppm.init h w m (fun _ _ -> Array.copy black))
      in
      exec_times cfg.timings_iter 
        (fun () -> runfilter cfg in_img out_img);
      Ppm.output_ppm cfg.oc !in_img;
    end;
    if cfg.time_simplify ||
       cfg.time_codegen ||
       cfg.time_bang ||
       cfg.time_exec then Timing.print_times ();
    finalize cfg;
  with Defs.Type_error msg->
    begin
      prerr_string "Type error: ";
      prerr_string msg;
      prerr_string "\n";
      exit 1;
    end
  | Parsing.Parse_error ->
    begin
      exit 1;
    end


