(*

  application: image filtering
  state: experimental
  revision: $Id: stagedinter.ml,v 1.3 2005/06/29 15:52:25 tobias Exp $
 
  Metaprogramming Project
  Authors: Christoph A. Herrmann and Tobias Langhammer
  University of Passau, Department of Informatics, Chair of Programming

 Copyright (C) 2004    University of Passau, Department of Informatics
 
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*)

(*
sed command for stripping MetaOCaml syntax from this file

sed -e "s/>\./)/g;s/\.</(/g;s/\.[!~]/ /g"\
    -e "s/C\(Bool\|Float\|Int\)/V\1/g"\
    -e "s/^.*__only_staged__.*$//"

*)

open Defs 
open Ppm


type 'a codevalue =              (* __only_staged__ *)
    CInt of ('a,int) code        (* __only_staged__ *)
  | CFloat of ('a,float) code    (* __only_staged__ *)
  | CBool of ('a,bool) code      (* __only_staged__ *)
  | CMatrix of ('a,float array array) code      (* __only_staged__ *)

(* val lift : value -> 'a codevalue *)
let lift = function
  | VInt   i -> CInt   .< i >.
  | VFloat f -> CFloat .< f >.
  | VBool  b -> CBool  .< b >.
  | VMatrix b -> CMatrix  .< b >. (* for testing without simplification *)

exception Inter_error of string
      

(* val unopfunCode : unOp -> 'a codevalue -> 'a codevalue *)
let unopfunCode op a =
  match (op,a) with
  | (NegI, CInt a)   -> CInt .< (- .~a) >.
  | (NegF, CFloat a) -> CFloat .< (-. .~a) >. 
  | (Not,  CBool a)  -> CBool .< not .~a >.
  | (AbsI, CInt a) -> CInt .< abs .~a >.
  | (AbsF, CFloat a) -> CFloat .< abs_float .~a >.
  | (Exp, CFloat a) -> CFloat .< exp .~a >.
  | (Log, CFloat a) -> CFloat .< log .~a >.
  | (Sqrt, CFloat a) -> CFloat .< sqrt .~a >.
  | (Sin, CFloat a) -> CFloat .< sin .~a >.
  | (Cos, CFloat a) -> CFloat .< cos .~a >.
  | (Tan, CFloat a) -> CFloat .< tan .~a >.
  | (ASin, CFloat a) -> CFloat .< asin .~a >.
  | (ACos, CFloat a) -> CFloat .< acos .~a >.
  | (ATan, CFloat a) -> CFloat .< atan .~a >.
  | _ -> raise (Inter_error 
     "in function 'unopfunCode': Cannot create code for given operand.")


(* val binopfunCode : binOp -> 'a codevalue -> 'a codevalue 
                                            -> 'a codevalue *)
let binopfunCode op a b = match (op,a,b) with
  | (AddI, CInt   a, CInt   b) -> CInt   .< .~a  +  .~b >.
  | (SubI, CInt   a, CInt   b) -> CInt   .< .~a  -  .~b >.
  | (AddF, CFloat a, CFloat b) -> CFloat .< .~a  +. .~b >.
  | (EqI,  CInt   a, CInt   b) -> CBool  .< .~a  =  .~b >.
  | (MulI, CInt a, CInt b) -> CInt .< .~a * .~b >.
  | (DivI, CInt a, CInt b) -> CInt .< .~a / .~b >.
  | (ModI, CInt a, CInt b) -> CInt .< .~a mod .~b >.
  | (SubF, CFloat a, CFloat b) -> CFloat .< .~a -. .~b >.
  | (MulF, CFloat a, CFloat b) -> CFloat .< .~a *. .~b >.
  | (DivF, CFloat a, CFloat b) -> CFloat .< .~a /. .~b >.
  | (PowF, CFloat a, CFloat b) -> CFloat .< .~a ** .~b >.
  | (EqB, CBool a, CBool b) -> CBool .< .~a = .~b >.
  | (UEqB, CBool a, CBool b) -> CBool .< .~a != .~b >.
  | (AndB, CBool a, CBool b) -> CBool .< .~a && .~b >.
  | (OrB, CBool a, CBool b) -> CBool .< .~a || .~b >.
  | (MinI, CInt a, CInt b) -> CInt .< min .~a .~b >.
  | (MaxI, CInt a, CInt b) -> CInt .< max .~a .~b >.
  | (MinF, CFloat a, CFloat b) -> CFloat .< min .~a .~b  >.
  | (MaxF, CFloat a, CFloat b) -> CFloat .< max .~a .~b  >.
  | (UEqI, CInt a, CInt b) -> CBool .< .~a != .~b >.
  | (LtI, CInt a, CInt b) -> CBool .< .~a < .~b >.
  | (LeI, CInt a, CInt b) -> CBool .< .~a <= .~b >.
  | (GtI, CInt a, CInt b) -> CBool .< .~a > .~b >.
  | (GeI, CInt a, CInt b) -> CBool .< .~a >= .~b >.
  | (EqF, CFloat a, CFloat b) -> CBool .< .~a = .~b >.
  | (UEqF, CFloat a, CFloat b) -> CBool .< .~a != .~b >.
  | (LtF, CFloat a, CFloat b) -> CBool .< .~a < .~b >.
  | (LeF, CFloat a, CFloat b) -> CBool .< .~a <= .~b >.
  | (GtF, CFloat a, CFloat b) -> CBool .< .~a > .~b >.
  | (GeF, CFloat a, CFloat b) -> CBool .< .~a >= .~b >.
  | _ -> raise (Inter_error 
     "in function 'binopfunCode': Cannot create code for given operands.")


let unCBool = function 
  | (CBool i) -> i
  | _ -> 
      raise (Inter_error "in funtion 'unCBool': Value is not bool.")
let unCInt = function 
  | (CInt i) -> i
  | _ -> 
      raise (Inter_error "in funtion 'unCInt': Value is not int.")
let unCFloat = function
  | (CFloat f) -> f
  | _ -> raise (Inter_error 
     "in funtion 'unCFloat': Value is not float.")
let unCMatrix = function
  | (CMatrix f) -> f
  | _ -> raise (Inter_error 
     "in funtion 'unCMatrix': Value is not a matrix.")

let saveUnCFloat = function 
  | (CFloat f) -> f
  | (CInt i) -> .< float_of_int .~i >.
  | _ -> raise (Inter_error 
     "in funtion 'saveUnCFloat': Value is neither int nor float.")
 

let rec iter n init step = 
  if n=0
     then init
     else let n'=n-1 in step .< n' >.  (iter n' init step)

(*
let dynForPartialUnroll copies low high init step =
            .< let s = ref .~init 
	       and index = ref .~low in
               while !index <= (.~high)-copies+1 do
	          .~(iter copies 
                          .< () >. 
                          (fun i code -> .< begin
                                             .~code;
                                             s := .~(step .< !index + .~i >. .<!s>.);
                                            end >.)  
                         );
                         index := !index + copies
	       done;
               while !index <= .~high do
	          s := .~(step .< !index >. .<!s>.);
		  index := !index + 1
               done;
	       !s
            >.

*)

(* val codegen_exp : int -> exp -> 'a codevalue environment 
     -> ppm_image -> 'a codevalue *)
let rec codegen_exp expr env source  =
  let { dtype=dtype; op=op; args=args } = expr in
  let codegen_all = 
    List.map (fun x -> codegen_exp x env source)
  in
  match op with
  | C v -> lift v
  | V s -> lookup_excp s env 
  | Image -> 
      let (row,col,ch) =
        match codegen_all args with
        | [CInt row; CInt col; CInt ch] -> (row,col,ch)
        | _ -> raise (Inter_error
        "in 'codegen_exp': Non-int arg for color operator. Bug in parsing?")
      in
      let rast = source.raster in
      let pnt = .< rast.(.~row).(.~col).(.~ch) >. in
      let maxv = float_of_int source.maxval in
      CFloat .< float_of_int .~pnt /. maxv >.
  | Int2Float -> let a = unCInt (head (codegen_all args)) in
                 CFloat .< float_of_int .~a >.
  | Floor     -> let a = unCFloat (head (codegen_all args)) in
                 CInt .< int_of_float .~a >.
  | UnOp  op -> let a = head (codegen_all args) 
                in unopfunCode op a  
  | BinOp op -> let (a,b) = head2 (codegen_all args) 
                in binopfunCode op a b  
  | If -> 
     let (c,a,b) = head3 args in
     let (c',a',b') = head3 (codegen_all args) in
     begin match a',b' with
       | CBool a,  CBool b  -> 
           CBool  .< if .~(unCBool c') then .~a else .~b >.
       | CInt a,   CInt b   -> 
           CInt   .< if .~(unCBool c') then .~a else .~b >.
       | CFloat a, CFloat b ->
           CFloat .< if .~(unCBool c') then .~a else .~b >.
       | _ -> raise (Inter_error
         "in 'codegen_exp': Invalid types for if-branches. Bug in parser?")
     end
  | Let s ->
     let (rhs, body) = head2 args in 
     let rhs' = codegen_exp rhs env source in
     let body_gen b = 
       let env' = extEnv (s,b) env in
       codegen_exp body env' source
     in
     begin match (rhs.dtype, body.dtype) with
       | (Int,Int)     -> 
           CInt .< let b = .~(unCInt rhs') in
             .~(unCInt (body_gen (CInt .<b>.))) >.
       | (Int,Float)   -> 
           CFloat .< let b = .~(unCInt rhs') in
             .~(unCFloat (body_gen (CInt .<b>.))) >.
       | (Int,Bool)   -> 
           CBool .< let b = .~(unCInt rhs') in
             .~(unCBool (body_gen (CInt .<b>.))) >.
       | (Float,Int)   ->
           CInt .< let b = .~(unCFloat rhs') in
             .~(unCInt (body_gen (CFloat .<b>.))) >.
       | (Float,Float) ->
           CFloat .< let b = .~(unCFloat rhs') in
             .~(unCFloat (body_gen (CFloat .<b>.))) >.
       | (Float,Bool) ->
           CBool .< let b = .~(unCFloat rhs') in
             .~(unCBool (body_gen (CFloat .<b>.))) >.
       | (Bool,Int)   ->
           CInt .< let b = .~(unCBool rhs') in
             .~(unCInt (body_gen (CBool .<b>.))) >.
       | (Bool,Float) ->
           CFloat .< let b = .~(unCBool rhs') in
             .~(unCFloat (body_gen (CBool .<b>.))) >.
       | (Bool,Bool) ->
           CBool .< let b = .~(unCBool rhs') in
             .~(unCBool (body_gen (CBool .<b>.))) >.
       | (Matrix,Int) ->
           CInt .< let b = .~(unCMatrix rhs') in
             .~(unCInt (body_gen (CMatrix .<b>.))) >.
       | (Matrix,Float) ->
           CFloat .< let b = .~(unCMatrix rhs') in
             .~(unCFloat (body_gen (CMatrix .<b>.))) >.
       | _ -> raise (Inter_error
(Printf.sprintf "in 'codegen_exp': Invalid types (rhs:%s,body:%s) for 'let'. Bug in parser?" (string_of_dtype rhs.dtype)
(string_of_dtype body.dtype)))
   end
  | Sum s ->
     let (lb,ub,body) = head3 args in
     let lb' = unCInt (codegen_exp lb env source) 
     and ub' = unCInt (codegen_exp ub env source) in
     let bodycode i = 
       let env' = extEnv (s,CInt i) env in
       codegen_exp body env' source 
     in
     let dynFor low high init step =
            .< let s = ref .~init in
               for i = .~low to .~high do
                  s := .~(step .<i>. .<!s>.)
               done;
               !s
            >.
     in
     begin match dtype with
     | Int   -> 
         let step i s = .< .~s +  .~(unCInt (bodycode i)) >. in
         CInt (dynFor lb' ub' .<0>. step) 
     | Float -> 
         let step i s = .< .~s +. .~(unCFloat (bodycode i)) >. in
         CFloat (dynFor lb' ub' .< 0. >. step) 
     | _ -> raise (Inter_error
                     "in 'codegen_exp': Invalid type for 'sum'. Bug in parser?")
     end 
 | IndexMatrix s ->
     let x,y = match head2 (codegen_all args) with
     | CInt x, CInt y -> x, y
     | _ -> raise (Inter_error
         "in 'codegen_exp': Invalid type for matrix indexes. Bug in parser?")
     in
     begin match lookup_excp s env with
     | CMatrix m ->
         CFloat .< (.~m).(.~x).(.~y) >.
     | _ -> raise (Inter_error ("variable '"^s^
                   "' is not of type matrix"))
     end

(* val codegen_top_exp :
    top_exp -> 'a codevalue environment ->
    ppm_image -> ('a, int list) code *)
let rec codegen_top_exp texp env image = 
  let { common_bind=bind; channels=exps } = texp in
  begin match bind with
    | [] ->
        let scale_to_int v = 
          let maxv = image.maxval in
          let maxvf = float_of_int maxv in
          .< let c = int_of_float (.~v *. maxvf ) in
             if c<0 then 0 else if c >maxv then maxv else c >.
        in
        let cg exp = scale_to_int 
           (saveUnCFloat (codegen_exp exp env image))
        in
        let channels = List.map cg exps in
        let cons_code hd tl = .< .~hd :: .~tl >. in
        List.fold_right cons_code channels .< [] >.
    | ((v,rhs)::bind') -> 
        let cexp =  { common_bind=bind'; channels=exps } in
        begin match codegen_exp rhs env image with
        | CInt r -> 
            .< let b = .~r in 
              .~(codegen_top_exp cexp (extEnv (v,CInt .<b>.) env) image) >.
        | CFloat r ->
            .< let b = .~r in 
              .~(codegen_top_exp cexp (extEnv (v,CFloat .<b>.) env) image) >.
        | CBool r -> 
            .< let b = .~r in 
              .~(codegen_top_exp cexp (extEnv (v,CBool .<b>.) env) image) >.
        | CMatrix r ->
            .< let b = .~r in 
              .~(codegen_top_exp cexp (extEnv (v,CMatrix .<b>.) env) image) >.
        end
  end

let parse funString = 
  Filt_parser.main Filt_lexer.token (Lexing.from_string funString)

let interp iter env cexp image =
  let (width,height) = image.width,image.height in
  let denv row col env =
    let stat_env = List.filter (fun (_,_,x) -> x != None) static_env in
    let stat_env' =
      List.map 
        (fun (x,_,y) -> 
          match y with
          | Some v -> (x,lift v)
          | None -> raise (Failure ""))
        stat_env
    in
    extEnvPairs
      (stat_env' @
        [ "iter", CInt .<iter>.;
          "height", CInt .< height>.;
          "width", CInt .< width>.;
          "row", CInt row;
          "col", CInt col
        ])
      env
  in
  .<
  fun (row,col) -> Array.of_list
    .~(codegen_top_exp cexp (denv .<row>.  .<col>. env) image)
  >.                     

let compile time_codegen time_bang iter ast image = 
  let (filter_code
  ) = Timing.timenew_cond time_codegen ("codegen")
     (fun () -> interp iter emptyEnv ast image) 
  in
  Timing.timenew_cond time_bang ("bang") (fun () -> .! filter_code)

