(*

  application: image filtering
  state: experimental
  revision: $Id: staticeval.ml,v 1.2 2005/06/21 11:39:41 tobias Exp $

  Metaprogramming Project
  Authors: Christoph A. Herrmann and Tobias Langhammer
  University of Passau, Department of Informatics, Chair of Programming

 Copyright (C) 2004    University of Passau, Department of Informatics

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*)


open Defs
open Ppm

let max_unrolling = ref (-1)

(* creating a unique name from a string *)
let unq = ref 0
let unique_name s =
  unq := !unq + 1;
  Printf.sprintf "%10d_%s" !unq s

let unopfun op a =
  match (op,a) with
    (NegI,VInt a) -> VInt (0 - a)
  | (NegF,VFloat a) -> VFloat (0.0 -. a)
  | (Not,VBool a) -> VBool (not a)
  | (AbsI, VInt a) -> VInt (abs a)
  | (AbsF, VFloat a) -> VFloat (if a<0.0 then -. a else a)
  | (Exp, VFloat a) -> VFloat (exp a)
  | (Log, VFloat a) -> VFloat (log a)
  | (Sqrt, VFloat a) -> VFloat (sqrt a)
  | (Sin, VFloat a) -> VFloat (sin a)
  | (Cos, VFloat a) -> VFloat (cos a)
  | (Tan, VFloat a) -> VFloat (tan a)
  | (ASin, VFloat a) -> VFloat (asin a)
  | (ACos, VFloat a) -> VFloat (acos a)
  | (ATan, VFloat a) -> VFloat (atan a)
  | _ -> raise (Type_error "in 'unopfun'")

let binopfun op a b =
    match (op,a,b) with
        (AddI, VInt a, VInt b) -> VInt (a+b)
      | (SubI, VInt a, VInt b) -> VInt (a-b)
      | (MulI, VInt a, VInt b) -> VInt (a*b)
      | (DivI, VInt a, VInt b) -> VInt (a/b)
      | (ModI, VInt a, VInt b) -> VInt (a mod b)
      | (AddF, VFloat a, VFloat b) -> VFloat (a+.b)
      | (SubF, VFloat a, VFloat b) -> VFloat (a-.b)
      | (MulF, VFloat a, VFloat b) -> VFloat (a*.b)
      | (DivF, VFloat a, VFloat b) -> VFloat (a/.b)
      | (PowF, VFloat a, VFloat b) -> VFloat (a**b)
      | (EqB, VBool a, VBool b) -> VBool (a=b)
      | (UEqB, VBool a, VBool b) -> VBool (a!=b)
      | (AndB, VBool a, VBool b) -> VBool (a&&b)
      | (OrB, VBool a, VBool b) -> VBool (a||b)
      | (MinI, VInt a, VInt b) -> VInt (min a b)
      | (MaxI, VInt a, VInt b) -> VInt (max a b)
      | (MinF, VFloat a, VFloat b) -> VFloat (min a b)
      | (MaxF, VFloat a, VFloat b) -> VFloat (max a b)
      | (EqI, VInt a, VInt b) -> VBool (a=b)
      | (UEqI, VInt a, VInt b) -> VBool (a!=b)
      | (LtI, VInt a, VInt b) -> VBool (a<b)
      | (LeI, VInt a, VInt b) -> VBool (a<=b)
      | (GtI, VInt a, VInt b) -> VBool (a>b)
      | (GeI, VInt a, VInt b) -> VBool (a>=b)
      | (EqF, VFloat a, VFloat b) -> VBool (a=b)
      | (UEqF, VFloat a, VFloat b) -> VBool (a!=b)
      | (LtF, VFloat a, VFloat b) -> VBool (a<b)
      | (LeF, VFloat a, VFloat b) -> VBool (a<=b)
      | (GtF, VFloat a, VFloat b) -> VBool (a>b)
      | (GeF, VFloat a, VFloat b) -> VBool (a>=b)
      | _ ->
          let s = Printf.sprintf "binopfun (%s,%s,%s)"
            (string_of_binOp op) (string_of_value a) (string_of_value b)
          in
          raise (Type_error s)





let rec substitute ((key, subst) as pair) expr =
  let { args=args } = expr in
  let substitute_all = List.map (substitute pair) in
  let exp_of_args new_args =
    create_exp expr.dtype expr.op new_args
  in
  match expr.op with
  | V s ->
      if s = key then subst
      else expr
  | C v -> exp_of_value v
  | Image | Int2Float | Floor | UnOp _ | BinOp _ | If ->
      exp_of_args (substitute_all args)
  | Let s ->
      let (rhs,body) = head2 args in
      let rhs' = substitute pair (head args) in
      if s = key then
        exp_of_args [rhs';body]
      else
        let body' = substitute pair body in
        exp_of_args (substitute_all [rhs';body'])
  | IndexMatrix s -> exp_of_args (substitute_all args)
  | Sum s ->
     let (lb,ub,body) = head3 args in
     let (lb',ub') = head2 (substitute_all [lb;ub]) in
     let body' = (if s=key then body else substitute pair body) in
     exp_of_args [lb';ub';body']


let zero_of t =
  match t with
  | Int -> exp_of_value (VInt 0)
  | Float -> exp_of_value (VFloat 0.0)
  | _ -> raise (Failure "zero_of")

let one_of t =
  match t with
  | Int -> exp_of_value (VInt 1)
  | Float -> exp_of_value (VFloat 1.0)
  | _ -> raise (Failure "one_of")

let negop_of = function
  | Int -> NegI
  | Float -> NegF
  | _ -> raise (Failure "negop_of")

let addop_of = function
  | Int -> AddI
  | Float -> AddF
  | _ -> raise (Failure "addop_of")

let subop_of = function
  | Int -> SubI
  | Float -> SubF
  | _ -> raise (Failure "subop_of")

let mulop_of = function
  | Int -> MulI
  | Float -> MulF
  | _ -> raise (Failure "mulop_of")

let divop_of = function
  | Int -> DivI
  | Float -> DivF
  | _ -> raise (Failure "divop_of")

let is_zero = function
  | VInt 0
  | VFloat 0.0 -> true
  | _ -> false

let is_one = function
  | VInt 1
  | VFloat 1.0 -> true
  | _ -> false

(* combines two expressions to a new expression which represents the
respective operation *)
let binop_exp op a b =
  let dtype = type_of_binop op in
  let combine a b = create_exp dtype (BinOp op) [a;b] in
  begin match op with
  | AddI | AddF ->
      begin match unC a,unC b with
      | Some av, Some bv ->
          exp_of_value (binopfun op av bv)
      | Some av, _ when is_zero av -> b
      | _, Some bv when is_zero bv -> a
      | _ -> combine a b
      end
  | SubI | SubF ->
      begin match unC a, unC b with
      | Some av, Some bv ->
          exp_of_value (binopfun op av bv)
      | Some av, _ when is_zero av ->
          create_exp dtype (UnOp (negop_of dtype)) [b]
      | _, Some bv when is_zero bv -> a
      | _ -> combine a b
      end
  | MulI | MulF ->
      begin match unC a, unC b with
      | Some av, Some bv ->
          exp_of_value (binopfun op av bv)
      | Some av, _ when is_zero av -> zero_of dtype
      | _, Some bv when is_zero bv -> zero_of dtype
      | Some av, _ when is_one av -> b
      | _, Some bv when is_one bv -> a
      | _ -> combine a b
      end
  | DivI | DivF ->
      begin match unC a,unC b with
      | Some av, Some bv ->
          exp_of_value (binopfun op av bv)
      | Some av, _ when is_zero av -> zero_of dtype
      | _, Some bv when is_one bv -> a
      | _ -> combine a b
      end
  | op ->
      begin match unC a,unC b with
      | Some av, Some bv ->
          exp_of_value (binopfun op av bv)
      | _ -> combine a b
      end
  end

let rec reduce_sum dtype lb ub iteration acc =
  if lb > ub then (lb, acc)
  else begin
    let add = binop_exp (addop_of dtype) in
    let iter = iteration lb in
    match unC iter with
    | Some c ->
        let acc' = add acc iter in
        reduce_sum dtype (lb+1) ub iteration acc'
    | None -> (lb, acc)
  end


(* simplify_exp: exp -> value environment -> exp *)
let rec simplify_exp expr env =
  let { dtype=dtype; op=op; args=args } = expr in
  let simplify_all = List.map (fun x -> simplify_exp x env) in
  let exp_of_args args =
    create_exp dtype op args
  in
  begin match op with
  | C v -> exp_of_value v
  | V s ->
      begin match lookup s env with
      | Some v -> exp_of_value v
      | None -> expr
      end
  | Image -> exp_of_args (simplify_all args)
  | Int2Float ->
      let subexp = head (simplify_all args) in
      begin match unC subexp with
      | Some (VInt i) ->
          exp_of_value (VFloat (float_of_int i))
      | Some _ -> raise (Type_error "Int2Float **must not happen**")
      | None -> exp_of_args [subexp]
      end
  | Floor ->
      let subexp = head (simplify_all args) in
      begin match unC subexp with
      | Some (VFloat i) ->
          exp_of_value (VInt (int_of_float i))
      | Some _ -> raise (Type_error "Floor **must not happen**")
      | None -> exp_of_args [subexp]
      end
  | UnOp op ->
      let subexp = head (simplify_all args) in
      begin match unC subexp with
      | Some a -> exp_of_value (unopfun op a)
      | None -> exp_of_args [subexp]
      end
  | BinOp op ->
      let aexp,bexp = head2 (simplify_all args) in
      binop_exp op aexp bexp
  | If ->
      let (cond,a,b) = head3 args in
      let cond' = simplify_exp cond env in
      begin match unC cond' with
      | Some (VBool c) ->
          if c then simplify_exp a env
          else simplify_exp b env
      | Some _ -> raise (Type_error "If **must not happen**")
      | None -> exp_of_args (cond'::simplify_all [a;b])
      end
  | Let s ->
      let (rhs,body) = head2 args in
      let rhs' = simplify_exp rhs env in
      begin match unC rhs' with
      | Some ((VMatrix _) as v) ->
          let body' = simplify_exp body (extEnv (s,v) env) in
          exp_of_args [rhs';body']
      | Some v -> simplify_exp body (extEnv (s,v) env)
      | None ->
          let body' = simplify_exp body env in
          begin match unC body' with
          | Some v -> exp_of_value v
          | None -> exp_of_args [rhs';body']
          end
      end
  | IndexMatrix s ->
      let (row,col) = head2 (simplify_all args) in
      begin match unC row, unC col, lookup s env with
      | Some (VInt r), Some (VInt c), Some (VMatrix m) ->
          exp_of_value (VFloat (m.(r).(c)))
      | _,_, Some (VMatrix m) -> exp_of_args [row;col]
      | _ -> raise (Type_error "IndexMatrix **must not happen**")
      end
  | Sum index ->
      begin
        let lower_orig, upper_orig, body_orig = head3 args in
        let lower, upper = head2 (simplify_all [lower_orig;upper_orig]) in

        (* try to reduce to const for static bounds *)
        let lower2, acc =
          match unC lower, unC upper with
          | Some (VInt lo), Some (VInt up) -> (* static bounds *)
              let iteration i =
                let env_iter = extEnv (index,VInt i) env in
                simplify_exp body_orig env_iter
              in
              let lo_new, acc =
                reduce_sum dtype lo up iteration (zero_of dtype) in
              exp_of_value (VInt lo_new),acc
          | _ -> lower, zero_of dtype
        in

        (* unroll the remaining  sum *)
        let body = simplify_exp body_orig env in
        let less a b =
          match unC a, unC b with
          | Some (VInt ac), Some (VInt bc) -> ac < bc
          | _ -> false
        in
        let unrolled_sum = 
          if less upper lower2 then
            zero_of dtype
          else begin
            let max_ur = (!max_unrolling / body.expand) - 1 in
            if max_ur <= 1 then
              exp_of_args [lower2;upper;body]
            else begin
              partial_unroll dtype max_ur index lower2 upper body env
            end
          end
        in
        binop_exp (addop_of dtype) acc unrolled_sum
      end
  end

(* val partial_unroll : 
    dtype -> int -> string -> exp 
     -> exp -> exp -> value environment 
     -> exp
*)
and partial_unroll dtype max_ur istr lo up body env =
  begin
    (*
   sum i from lo to up of body(i)

       transformed to

   sum k from 0 to (up - lo + 1) div N - 1 of
     let b = lo + k * N in
     sum s from 0 to N-1 of
       body(b + s)
   +
   sum t from ((up - lo + 1) div N) * N + lo to up of
     body(t)

    *)

    (* helpers to abbreviate Node construction... *)
    let ( -: ) = binop_exp SubI in
    let ( +: ) = binop_exp AddI in
    let ( /: ) = binop_exp DivI in
    let ( *: ) = binop_exp MulI in
    let n = exp_of_value (VInt max_ur) in
    let kstr = unique_name "k" in
    let sstr = unique_name "s" in
    let tstr = unique_name "t" in
    let bstr = unique_name "b" in
    let var_of_str s = create_exp Int (V s) [] in
    let k = var_of_str kstr in
    let s = var_of_str sstr in
    let t = var_of_str tstr in
    let b = var_of_str bstr in

    (* complete unrolling of a loop *)
    let rec unroll ix lb ub acc body =
      if ub < lb then acc
      else
        let body_instance =
          let env' = extEnv (ix,VInt lb) env in
          simplify_exp body env' in
        let acc' = binop_exp (addop_of dtype) acc body_instance in
        unroll ix (lb+1) ub acc' body
    in

    let no_of_blocks = ((up -: lo) +: one_of Int) /: n in

    (* construct sum of k *)
    let up_k = (no_of_blocks -: one_of Int) in
    let sum_k =
      match unC up_k with
      | Some (VInt u) when u < 0 -> zero_of Int
      | _ ->
        begin
          (* construct and unroll sum of s *)
          let up_s = exp_of_value (VInt (max_ur-1)) in
          let access = s +: b in
          let body_s = substitute (istr,access) body in
          let unrolled_s = unroll sstr 0 (max_ur-1) (zero_of Int) body_s in
          let body_k =
            create_exp dtype (Let bstr) [(k *: n +: lo); unrolled_s]
          in
          match unC up_k with
          | Some (VInt 0) ->
              simplify_exp body_k (extEnv (kstr,VInt 0) env)
          | _ ->
              create_exp dtype (Sum kstr) [zero_of Int; up_k; body_k]
        end
    in

    (* construct sum of t *)
    let lo_t = simplify_exp (no_of_blocks *: n +: lo) env in
    let body_t = substitute (istr,t) body in
    let sum_t =
      match unC lo_t, unC up with
      | Some (VInt lt), Some (VInt ut) ->
          unroll tstr lt ut (zero_of Int) body_t
      | _,_->
          create_exp dtype (Sum tstr) [lo_t;up;body_t]
    in
    binop_exp (addop_of dtype) sum_k sum_t
  end



(* val simplify_top_exp : top_exp -> value environment -> top_exp *)
let rec simplify_top_exp { common_bind=bind; channels=exps } env =
  match bind with
  | [] -> { common_bind=[]; 
            channels=List.map (fun e -> simplify_exp e env) exps }
  | ((k,rhs)::bindrest) -> 
      let rhs' = simplify_exp rhs env in
      begin match unC rhs' with
      | Some v -> simplify_top_exp 
                    { common_bind=bindrest; channels=exps }
                    (extEnv (k,v) env)
      | None -> 
          let { common_bind=b; channels=c } =
            simplify_top_exp
              { common_bind=bindrest; channels=exps }
              env
          in
          if List.exists (fun exp -> unC exp == None) c then
            { common_bind=(k,rhs')::b; channels=c }
          else
            { common_bind=b; channels=c }
      end
  
let simpl maxur ast width height iter =
  max_unrolling := maxur;
  let complete_static_env defs =
    List.map 
      (fun (x,_,y) -> 
         match y with 
         | Some v -> (x,v)
         | None -> (x,List.assoc x defs))
      static_env 
  in
  let env = newEnv (complete_static_env
           ["iter",       VInt iter;
            "width",      VInt width;
            "height",     VInt height ])
  in
  let { common_bind=bind; channels=exps } = ast in

  let simp = simplify_top_exp ast env
  in simp


let parse funString =
  Filt_parser.main Filt_lexer.token (Lexing.from_string funString)



